import boto3
import botocore
from botocore.exceptions import ClientError
from boto3.s3.transfer import TransferConfig
from io import StringIO, BytesIO
import os
import json
import larry.sts as sts
import uuid
import urllib.request
import urllib.parse
import larry.utils
from zipfile import ZipFile

# Local S3 resource object
resource = None
# A local instance of the boto3 session to use
__session = boto3.session.Session()


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws__session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws__session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session, resource
    __session = boto_session if boto_session is not None else boto3.session.Session(**larry.utils.copy_non_null_keys(locals()))
    sts.set_session(boto_session=__session)
    resource = __session.resource('s3')


def delete_object(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Deletes the object defined by the bucket/key pair (or uri).
    :param bucket: The S3 bucket
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: Dict containing the boto3 response
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    return s3_resource.Bucket(bucket).Object(key=key).delete()


def get_object(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Performs a 'get' of the object defined by the bucket/key pair (or uri).
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: Dict containing the Body of the object and associated attributes
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    return s3_resource.Bucket(bucket).Object(key=key).get()


def get_object_size(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Returns the content_length of an S3 object.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: Size in bytes
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    return s3_resource.Bucket(bucket).Object(key=key).content_length


def read_object(bucket=None, key=None, uri=None, amt=None, s3_resource=None):
    """
    Performs a 'get' of the object defined by the bucket/key pair (or uri)
    and then performs a 'read' of the Body of that object.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param amt: The max amount of bytes to read from the object. All data is read if omitted.
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The bytes contained in the object
    """
    s3_resource = s3_resource if s3_resource else resource
    return get_object(bucket, key, uri, s3_resource=s3_resource)['Body'].read(amt)


def read_dict(bucket=None, key=None, uri=None, encoding='utf-8', s3_resource=None):
    """
    Reads in the s3 object defined by the bucket/key pair (or uri) and
    loads the json contents into a dict.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param encoding: The charset to use when decoding the object bytes, utf-8 by default
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: A dict representation of the json contained in the object
    """
    s3_resource = s3_resource if s3_resource else resource
    return json.loads(read_object(bucket, key, uri, s3_resource=s3_resource).decode(encoding),
                      object_hook=larry.utils.JSONDecoder)


def read_str(bucket=None, key=None, uri=None, encoding='utf-8', s3_resource=None):
    """
    Reads in the s3 object defined by the bucket/key pair (or uri) and
    decodes it to text.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param encoding: The charset to use when decoding the object bytes, utf-8 by default
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The contents of the object as a string
    """
    s3_resource = s3_resource if s3_resource else resource
    return read_object(bucket, key, uri, s3_resource=s3_resource).decode(encoding)


def read_list_of_dict(bucket=None, key=None, uri=None, encoding='utf-8', newline='\n', s3_resource=None):
    """
    Reads in the s3 object defined by the bucket/key pair (or uri) and
    loads the JSON Lines data into a list of dict objects.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param encoding: The charset to use when decoding the object bytes, utf-8 by default
    :param newline: The line separator to use when reading in the object, \n by default
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The contents of the object as a list of dict objects
    """
    s3_resource = s3_resource if s3_resource else resource
    obj = read_object(bucket, key, uri, s3_resource=s3_resource)
    lines = obj.decode(encoding).split(newline)
    records = []
    for line in lines:
        if len(line) > 0:
            record = json.loads(line, object_hook=larry.utils.JSONDecoder)
            records.append(record)
    return records


def read_list_of_str(bucket=None, key=None, uri=None, encoding='utf-8', newline='\n', s3_resource=None):
    """
    Reads in the s3 object defined by the bucket/key pair (or uri) and
    loads the JSON Lines data into a list of dict objects.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param encoding: The charset to use when decoding the object bytes, utf-8 by default
    :param newline: The line separator to use when reading in the object, \n by default
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The contents of the object as a list of dict objects
    """
    s3_resource = s3_resource if s3_resource else resource
    obj = read_object(bucket, key, uri, s3_resource=s3_resource)
    lines = obj.decode(encoding).split(newline)
    records = []
    for line in lines:
        if len(line) > 0:
            records.append(line)
    return records


def read_pillow_image(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Reads in the s3 object defined by the bucket/key pair (or uri) and
    loads it into a Pillow image object
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The contents of the object as a Pillow image object
    """
    s3_resource = s3_resource if s3_resource else resource
    try:
        from PIL import Image
        return Image.open(BytesIO(read_object(bucket, key, uri, s3_resource=s3_resource)))
    except ImportError as e:
        # Simply raise the ImportError to let the user know this requires Pillow to function
        raise e


def write(body, bucket=None, key=None, uri=None, acl=None, content_type=None, s3_resource=None):
    """
    Write an object to the bucket/key pair (or uri).
    :param body: Data to write
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param acl: The canned ACL to apply to the object
    :param content_type: A standard MIME type describing the format of the object data
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    obj = s3_resource.Bucket(bucket).Object(key=key)
    if acl and content_type:
        obj.put(Body=body, ACL=acl, ContentType=content_type)
    elif acl:
        obj.put(Body=body, ACL=acl)
    elif content_type:
        obj.put(Body=body, ContentType=content_type)
    else:
        obj.put(Body=body)
    return compose_uri(bucket, key)


def write_temp_object(value, prefix, acl=None, s3_resource=None, bucket_identifier=None, region=None,
                      bucket=None):
    """
    Write an object to a temp bucket with a unique UUID.
    :param value: Object to write to S3
    :param prefix: Prefix to attach ahead of the UUID as the key
    :param acl: The canned ACL to apply to the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :param bucket_identifier: The identifier to attach to the temp bucket that will be used for writing to s3, typically
    the account id (from STS) for the account being used
    :param region: The s3 region to store the data in
    :param bucket: The bucket to use instead of creating/using a temp bucket
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource
    if bucket is None:
        bucket = get_temp_bucket(region=region, bucket_identifier=bucket_identifier, s3_resource=s3_resource)
    key = prefix + str(uuid.uuid4())
    return write_object(value, bucket=bucket, key=key, acl=acl, s3_resource=s3_resource)


def write_object(value, bucket=None, key=None,
                 uri=None,
                 acl=None,
                 newline='\n',
                 json_default=str,
                 content_type=None,
                 s3_resource=None):
    """
    Write an object to the bucket/key pair (or uri), converting the python
    object to an appropriate format to write to file.
    :param value: Object to write to S3
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param acl: The canned ACL to apply to the object
    :param newline: Character(s) to use as a newline for list objects
    :param json_default: default function for rendering data types
    :param content_type: Content type to apply to the file, if not present a suggested type will be applied
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource
    extension_types = {
        'css': 'text/css',
        'html': 'text/html',
        'xhtml': 'text/html',
        'htm': 'text/html',
        'xml': 'text/xml',
        'csv': 'text/csv',
        'txt': 'text/plain',
        'png': 'image/png',
        'jpeg': 'image/jpeg',
        'jpg': 'image/jpeg',
        'gif': 'image/gif',
        'jsonl': 'application/x-jsonlines',
        'json': 'application/json',
        'js': 'application/javascript',
        'zip': 'application/zip',
        'pdf': 'application/pdf',
        'sql': 'application/sql'
    }
    if uri:
        (bucket, key) = decompose_uri(uri)
    extension = key.split('.')[-1]
    # JSON
    if type(value) is dict:
        if content_type is None:
            content_type = 'application/json'
        return write(json.dumps(value, default=json_default, cls=larry.utils.JSONEncoder), bucket, key, uri,
                     acl, content_type=content_type, s3_resource=s3_resource)
    # Text
    elif type(value) is str:
        if content_type is None:
            content_type = extension_types.get(extension, 'text/plain')
        return write(value, bucket, key, uri, acl, content_type=content_type, s3_resource=s3_resource)
    elif type(value) is list:
        if content_type is None:
            content_type = extension_types.get(extension, 'text/plain')
        buff = StringIO()
        for row in value:
            if type(row) is dict:
                buff.write(json.dumps(row, default=json_default, cls=larry.utils.JSONEncoder) + newline)
            else:
                buff.write(str(row) + newline)
        return write(buff.getvalue(), bucket, key, uri, acl, content_type=content_type, s3_resource=s3_resource)
    elif value is None:
        return write('', bucket, key, uri, acl, s3_resource=s3_resource, content_type=content_type)
    else:
        # try to write it as an image
        try:
            buff = BytesIO()
            fmt = 'PNG' if value.format is None else value.format
            value.save(buff, fmt)
            buff.seek(0)
            if content_type is None:
                content_type = extension_types.get(extension, extension_types.get(fmt.lower(), 'text/plain'))
            return write(buff, bucket, key, uri, content_type=content_type, s3_resource=s3_resource)
        except Exception:
            return write(value, bucket, key, uri, acl, content_type=content_type, s3_resource=s3_resource)


def write_pillow_image(image, image_format, bucket=None, key=None, uri=None, s3_resource=None):
    """
    Write an image to the bucket/key pair (or uri).
    :param image: The image to write to S3
    :param image_format: The format of the image (png, jpeg, etc)
    :param bucket: The S3 bucket for the object
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource
    buff = BytesIO()
    image.save(buff, image_format)
    buff.seek(0)
    return write(buff, bucket, key, uri, s3_resource=s3_resource)


def write_as_csv(rows, bucket=None, key=None, uri=None, acl=None, delimiter=',', columns=None, headers=None,
                 s3_resource=None):
    """
    Write an object to the bucket/key pair (or uri), converting the python
    object to an appropriate format to write to file.
    :param rows: List of data to write, rows can be of type list, dict or str
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param acl: The canned ACL to apply to the object
    :param delimiter: Column delimiter to use, ',' by default
    :param columns: The columns to write out from the source rows, dict keys or list indexes
    :param headers: Headers to add to the output
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource

    def _array_to_string(_row, _delimiter, _indices=None):
        if _indices is None:
            _indices = range(len(_row))
        _line = ''
        for x in _indices:
            _line = str(row[x]) if x == 0 else _line + _delimiter + str(row[x])
        return _line

    buff = StringIO()

    # empty
    if rows is None or len(rows) == 0:
        if headers:
            buff.write(_array_to_string(headers, delimiter) + "\n")
        buff.write('')

    # list
    elif type(rows[0]) is list:
        indices = columns if columns else None
        if headers:
            buff.write(_array_to_string(headers, delimiter) + "\n")
        for row in rows:
            buff.write(_array_to_string(row, delimiter, indices) + "\n")

    # dict
    elif type(rows[0]) is dict:
        keys = columns if columns else rows[0].keys()
        buff.write(_array_to_string(headers if headers else keys, delimiter) + "\n")

        for row in rows:
            line = ''
            for i, k in enumerate(keys):
                value = '' if row.get(k) is None else str(row.get(k))
                line = value if i == 0 else line + delimiter + value
            buff.write(line + "\n")

    # string
    elif type(rows[0]) is str:
        buff.writelines(rows)
    else:
        raise Exception('Invalid input')
    return write(buff.getvalue(), bucket, key, uri, acl, s3_resource=s3_resource)


def rename_object(old_bucket_name, old_key, new_bucket_name, new_key, s3_resource=None):
    """
    Renames an object in S3.
    :param old_bucket_name: Source bucket
    :param old_key: Source key
    :param new_bucket_name: Target bucket
    :param new_key: Target key
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: None
    """
    s3_resource = s3_resource if s3_resource else resource
    copy_source = {
        'Bucket': old_bucket_name,
        'Key': old_key
    }
    s3_resource.meta.client.copy(copy_source, new_bucket_name, new_key)
    s3_resource.meta.client.delete_object(Bucket=old_bucket_name, Key=old_key)


def object_exists(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Checks to see if an object with the given bucket/key (or uri) exists.
    :param bucket: The S3 bucket for the object
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: True if the key exists, if not, False
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    try:
        s3_resource.Object(bucket, key).load()
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            return False
        else:
            raise e
    return True


def _find_largest_common_prefix(values):
    """
    Searches through a list of values to find the longest possible common prefix amongst them. Useful for optimizing
    more costly searches. Supports lists of strings or tuples. If tuples are used, the first value is assumed to be
    the value to search on.
    :param values: List of values (strings or tuples containing a string in the first position)
    :return: String prefix common to all values
    """
    if type(values[0]) is tuple:
        prefix, *_ = values[0]
    else:
        prefix = values[0]

    for value in values:
        key = value[0] if type(value) is tuple else value
        while key[:len(prefix)] != prefix and len(prefix) > 0:
            prefix = prefix[:-1]
    return prefix


def find_keys_not_present(bucket, keys=None, uris=None, s3_resource=None):
    """
    Searches an S3 bucket for a list of keys and returns any that cannot be found.
    :param bucket: The S3 bucket to search
    :param keys: A list of keys to search for (strings or tuples containing a string in the first position)
    :param uris: A list of S3 URIs to search for (strings or tuples containing a string in the first position)
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: A list of keys that were not found (strings or tuples based on the input values)
    """
    s3_resource = s3_resource if s3_resource else resource

    # If URIs are passed, convert to a list of keys to use for the search
    if uris:
        keys = []
        for value in uris:
            if type(value) is tuple:
                uri, *z = value
                b, key = decompose_uri(uri)
                keys.append(tuple([key]) + tuple(z))
            else:
                b, key = decompose_uri(value)
                keys.append(key)

    # Find the longest common prefix to use as the search term
    prefix = _find_largest_common_prefix(keys)

    # Get a list of all keys in the bucket that match the prefix
    bucket_obj = s3_resource.Bucket(bucket)
    all_keys = []
    for obj in bucket_obj.objects.filter(Prefix=prefix):
        all_keys.append(obj.key)

    # Search for any keys that can't be found
    not_found = []
    for value in keys:
        key = value[0] if type(value) is tuple else value
        if key not in all_keys:
            not_found.append(value)
    return not_found


def decompose_uri(uri):
    """
    Decompose an S3 URI into a bucket and key
    :param uri: S3 URI
    :return: Tuple containing a bucket and key
    """
    bucket_name = get_bucket_name(uri)
    return bucket_name, get_bucket_key(bucket_name, uri)


def get_bucket_name(uri):
    """
    Retrieve the bucket portion from an S3 URI
    :param uri: S3 URI
    :return: Bucket name
    """
    return uri.split('/')[2]


def get_bucket_key(bucket_name, uri):
    """
    Retrieves the key portion of an S3 URI
    :param bucket_name: S3 bucket name
    :param uri: S3 URI
    :return: Key value
    """
    pos = uri.find(bucket_name) + len(bucket_name) + 1
    return uri[pos:]


def compose_uri(bucket, key):
    """
    Compose a bucket and key into an S3 URI
    :param bucket: Bucket name
    :param key: Object key
    :return: S3 URI string
    """
    return "s3://{}/{}".format(bucket, key)


def list_objects(bucket=None, prefix=None, uri=None, include_empty_keys=False, s3_resource=None):
    """
    Returns a list of the object keys in the provided bucket that begin with the provided prefix.
    :param bucket: The S3 bucket to query
    :param prefix: The key prefix to use in searching the bucket
    :param uri: An s3:// path containing the bucket and prefix
    :param include_empty_keys: True if you want to include keys associated with objects of size=0
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: A generator of object keys
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, prefix) = decompose_uri(uri)
    paginator = s3_resource.meta.client.get_paginator('list_objects')
    operation_parameters = {'Bucket': bucket, 'Prefix': prefix}
    page_iterator = paginator.paginate(**operation_parameters)
    for page in page_iterator:
        for obj in page.get('Contents', []):
            if obj['Size'] > 0 or include_empty_keys:
                yield obj['Key']


def fetch(url, bucket=None, key=None, uri=None, s3_resource=None):
    """
    Retrieves the data defined by a URL to an S3 location.
    :param url: URL to retrieve
    :param bucket: The S3 bucket for the object
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URI of the object written to S3
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    try:
        with urllib.request.urlopen(url) as response:
            return write_object(response.read(), bucket=bucket, key=key, s3_resource=s3_resource)
    except Exception as e:
        print('Failed to retrieve {} due to {}'.format(url, e))


def download(directory, bucket=None, key=None, uri=None, use_threads=True, s3_resource=None):
    """
    Downloads the an S3 object to a directory on the local file system.
    :param directory: The directory to download the object to
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param use_threads: Enables the use_threads transfer config
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: Path of the local file
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    config = TransferConfig(use_threads=use_threads)
    s3_object_local = os.path.join(directory, key.split('/')[-1])
    try:
        s3_resource.Bucket(bucket).download_file(key, s3_object_local, Config=config)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            print("The object does not exist.")
        else:
            raise
    return s3_object_local


def upload(file_name, bucket=None, key=None, uri=None, s3_resource=None):
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    s3_resource.Bucket(bucket).upload_file(file_name, key)
    return compose_uri(bucket, key)


def download_to_temp(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Downloads the an S3 object to a temp directory on the local file system.
    :param bucket: The S3 bucket for object to retrieve
    :param key: The key of the object to be retrieved from the bucket
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: Path of the local file
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    temp_dir = _create_temp_dir()
    file = os.path.join(temp_dir, key.split('/')[-1])
    if not os.path.isfile(file):
        print('starting download')
        download(temp_dir, bucket, key, use_threads=True, s3_resource=s3_resource)
        print('download complete')
    return file


def _create_temp_dir():
    """
    Creates a temp directory in the current path.
    :return: The path of the temp directory
    """
    _temp_dir = os.getcwd() + "/temp"
    if not os.path.isdir(_temp_dir):
        os.makedirs(_temp_dir)
    return _temp_dir


def make_public(bucket=None, key=None, uri=None, s3_resource=None):
    """
    Makes the object defined by the bucket/key pair (or uri) public.
    :param bucket: The S3 bucket for object
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :return: The URL of the object
    """
    s3_resource = s3_resource if s3_resource else resource
    if uri:
        (bucket, key) = decompose_uri(uri)
    s3_resource.meta.client.put_object_acl(Bucket=bucket, Key=key, ACL='public-read')
    return get_public_url(bucket=bucket, key=key)


def get_public_url(bucket=None, key=None, uri=None):
    """
    Returns the public URL of an S3 object (assuming it's public).
    :param bucket: The S3 bucket for object
    :param key: The key of the object
    :param uri: An s3:// path containing the bucket and key of the object
    :return: The URL of the object
    """
    if uri:
        (bucket, key) = decompose_uri(uri)
    return 'https://{}.s3.amazonaws.com/{}'.format(bucket, urllib.parse.quote(key))


def create_bucket(bucket, acl='private', region=__session.region_name, s3_resource=None):
    s3_resource = s3_resource if s3_resource else resource
    bucket_obj = s3_resource.Bucket(bucket)
    bucket_obj.load()
    if bucket_obj.creation_date is None:
        bucket_obj.create(ACL=acl, CreateBucketConfiguration={'LocationConstraint': region})
        bucket_obj.wait_until_exists()
    return bucket_obj


def delete_bucket(bucket, s3_resource=None):
    s3_resource = s3_resource if s3_resource else resource
    bucket_obj = s3_resource.Bucket(bucket)
    bucket_obj.delete()


def get_temp_bucket(region=None, s3_resource=None, bucket_identifier=None):
    """
    Create a bucket that will be used as temp storage for larry commands.
    The bucket will be created in the region associated with the current session
    using a name based on the current session account id and region.
    :param region: Region to locate the temp bucket
    :param s3_resource: Boto3 resource to use if you don't wish to use the default resource
    :param bucket_identifier: The bucket identifier to use as a unique identifier for the bucket, defaults to the
    account id associated with the session
    :return: The name of the created bucket
    """
    if region is None:
        region = __session.region_name
    if bucket_identifier is None:
        bucket_identifier = sts.account_id()
    bucket = '{}-larry-{}'.format(bucket_identifier, region)
    create_bucket(bucket, region=region, s3_resource=s3_resource)
    return bucket


# TODO: add filter parameter
# TODO: rationalize the list params
def download_to_zip(file, bucket, prefix=None, prefixes=None):
    if prefix:
        prefixes = [prefix]
    with ZipFile(file, 'w') as zfile:
        for prefix in prefixes:
            for key in list_objects(bucket, prefix):
                zfile.writestr(urllib.parse.quote(key), data=read_object(bucket, key))


def file_name_portion(uri):
    file = decompose_uri(uri)[1].split('/')[-1]
    return file[:file.rfind('.')]



import larry.mturk
import larry.utils
import collections


class Assignment(collections.UserDict):
    def __init__(self, data, mturk_client=None):
        self.__client = mturk_client
        collections.UserDict.__init__(self)
        if isinstance(data, str):
            assignment, hit = larry.mturk._get_assignment(data, mturk_client)
            self.update(assignment)
        else:
            self.update(data)
            self._parse_datetime_values()
        if isinstance(self['Answer'], str):
            self['Answer'] = larry.mturk.parse_answers(self['Answer'])

    def _parse_datetime_values(self):
        for key in ['AutoApprovalTime', 'AcceptTime', 'SubmitTime', 'ApprovalTime', 'RejectionTime', 'Deadline']:
            try:
                if key in self and isinstance(self[key], str):
                    self[key] = larry.utils.parse_date(self[key])
            except ValueError:
                pass

    def __repr__(self):
        return "{}('{}')".format(type(self).__name__, self.assignment_id)

    def __str__(self):
        return "<{}: {}>".format(self.assignment_id, self.status)

    def refresh(self):
        self.update(larry.mturk._get_assignment(self.assignment_id, self.__client)[0])

    @property
    def assignment_id(self):
        return self['AssignmentId']

    @property
    def worker_id(self):
        return self['WorkerId']

    @property
    def hit_id(self):
        return self['HITId']

    @property
    def status(self):
        return self['AssignmentStatus'] # 'Submitted' | 'Approved' | 'Rejected'

    @property
    def auto_approval_time(self):
        return self.get('AutoApprovalTime', None)

    @property
    def accept_time(self):
        return self.get('AcceptTime', None)

    @property
    def submit_time(self):
        return self.get('SubmitTime', None)

    @property
    def approval_time(self):
        return self.get('ApprovalTime', None)

    @property
    def rejection_time(self):
        return self.get('RejectionTime', None)

    @property
    def work_time(self):
        return self.submit_time-self.accept_time if self.submit_time and self.accept_time else None

    @property
    def deadline(self):
        return self.get('Deadline', None)

    @property
    def answer(self):
        return self.get('Answer', None)

    @property
    def requester_feedback(self):
        return self.get('RequesterFeedback', None)



import json
import larry.s3 as s3
import larry.utils
import boto3
import os
from botocore.exceptions import ClientError
from larry.utils.image import scale_image_to_size


client = None
# A local instance of the boto3 session to use
__session = boto3.session.Session()


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws__session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws__session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session, client
    __session = boto_session if boto_session is not None else boto3.session.Session(**larry.utils.copy_non_null_keys(locals()))
    __client = __session.client('sagemaker')


def build_pre_response(event, log_details=True, input_attribute='taskObject'):
    if log_details:
        print(event)
    source = event['dataObject'].get('source', event['dataObject'].get('source-ref'))
    if source is None:
        print('No source or source-ref value found')
        return {}
    if log_details:
        print('Task object is: {}'.format(source))
    response = {'taskInput': {}}
    response['taskInput'][input_attribute] = source
    if log_details:
        print('Response is {}'.format(json.dumps(response)))
    return response


def build_pre_response_from_object(event, log_details=True, input_attribute='taskObject', s3_resource=None):
    if log_details:
        print(event)
    source_ref = event['dataObject'].get('source-ref')
    if source_ref:
        if s3_resource:
            value = s3.read_dict(uri=source_ref, s3_resource=s3_resource)
        else:
            value = s3.read_dict(uri=source_ref)
    else:
        source = event['dataObject'].get('source')
        if source is None:
            print('No source or source-ref value found')
            return {}
        else:
            if type(source) is str:
                value = json.loads(source)
            else:
                value = source
    if log_details:
        print('Task object is: {}'.format(json.dumps(value)))
    response = {'taskInput': {}}
    response['taskInput'][input_attribute] = value
    if log_details:
        print('Response is {}'.format(json.dumps(response)))
    return response


def build_consolidation_response(event, log_details=True):
    if log_details:
        print(json.dumps(event))
    payload = get_payload(event)
    if log_details:
        print('Payload: {}'.format(json.dumps(payload)))

    consolidated_response = []
    for dataset in payload:
        responses = extract_worker_responses(dataset['annotations'])
        consolidated_response.append({
            'datasetObjectId': dataset['datasetObjectId'],
            'consolidatedAnnotation': {
                'content': {
                    event['labelAttributeName']: {
                        'responses': responses
                    }
                }
            }
        })
    if log_details:
        print('Consolidated response: {}'.format(json.dumps(consolidated_response)))
    return consolidated_response


def extract_worker_responses(annotations):
    responses = []
    for annotation in annotations:
        response = json.loads(annotation['annotationData']['content'])
        if 'annotatedResult' in response:
            response = response['annotatedResult']

        responses.append({
            'workerId': annotation['workerId'],
            'annotation': response
        })
    return responses


def get_payload(event):
    if 'payload' in event:
        return s3.read_dict(uri=event['payload']['s3Uri'])
    else:
        return event.get('test_payload', [])


def _input_config(manifest_uri, free_of_pii=False, free_of_adult_content=True):
    config = {
        'DataSource': {
            'S3DataSource': {
                'ManifestS3Uri': manifest_uri
            }
        }
    }
    content_classifiers = []
    if free_of_adult_content:
        content_classifiers.append('FreeOfAdultContent')
    if free_of_pii:
        content_classifiers.append('FreeOfPersonallyIdentifiableInformation')
    if len(content_classifiers) > 0:
        config['DataAttributes'] = {'ContentClassifiers': content_classifiers}
    return config


def _output_config(output_uri, kms_key=None):
    config = {
        'S3OutputPath': output_uri
    }
    if kms_key:
        config['KmsKeyId'] = kms_key
    return config


def build_human_task_config(template_uri, pre_lambda_arn, consolidation_lambda_arn, title, description, workers=1,
                            public=False, reward_in_cents=None, workteam_arn=None, time_limit=300, lifetime=345600,
                            max_concurrent_tasks=None, keywords=None, region=None):
    region = __session.region_name if region is None else region
    config = {
        'UiConfig': {
            'UiTemplateS3Uri': template_uri
        },
        'PreHumanTaskLambdaArn': pre_lambda_arn,
        'TaskTitle': title,
        'TaskDescription': description,
        'NumberOfHumanWorkersPerDataObject': workers,
        'TaskTimeLimitInSeconds': time_limit,
        'TaskAvailabilityLifetimeInSeconds': lifetime,
        'AnnotationConsolidationConfig': {
            'AnnotationConsolidationLambdaArn': consolidation_lambda_arn
        }
    }
    if public:
        config['WorkteamArn'] = 'arn:aws:sagemaker:{}:394669845002:workteam/public-crowd/default'.format(region)
        if reward_in_cents is None:
            raise Exception('You must provide a reward amount for a public labeling job')
        else:
            config['PublicWorkforceTaskPrice'] = {
                'AmountInUsd': {
                    'Dollars': int(reward_in_cents // 100),
                    'Cents': int(reward_in_cents),
                    'TenthFractionsOfACent': round((reward_in_cents % 1) * 10)
                }
            }
    elif workteam_arn is not None:
        config['WorkteamArn'] = workteam_arn
    else:
        raise Exception('Labeling job must be public or have a workteam ARN')
    if keywords:
        config['TaskKeywords'] = keywords
    if max_concurrent_tasks:
        config['MaxConcurrentTaskCount'] = max_concurrent_tasks
    return config


def build_stopping_conditions(max_human_labeled_object_count=None, max_percentage_labeled=None):
    if max_human_labeled_object_count is None or max_percentage_labeled is None:
        return None
    else:
        config = {}
        if max_human_labeled_object_count:
            config['MaxHumanLabeledObjectCount'] = max_human_labeled_object_count
        if max_percentage_labeled:
            config['MaxPercentageOfInputDatasetLabeled'] = max_percentage_labeled
        return config


def build_algorithms_config(algorithm_specification_arn, initial_active_learning_model_arn=None, kms_key=None):
    if algorithm_specification_arn is None:
        return None
    config = {
        'LabelingJobAlgorithmSpecificationArn': algorithm_specification_arn
    }
    if initial_active_learning_model_arn:
        config['InitialActiveLearningModelArn'] = initial_active_learning_model_arn
    if kms_key:
        config['LabelingJobResourceConfig'] = {'VolumeKmsKeyId': kms_key}
    return config


def create_job(name,
               manifest_uri,
               output_uri,
               role_arn,
               task_config,
               category_config_uri=None,
               label_attribute_name=None,
               free_of_pii=False,
               free_of_adult_content=True,
               algorithms_config=None,
               stopping_conditions=None,
               sagemaker_client=None):
    sagemaker_client = sagemaker_client if sagemaker_client else client
    if label_attribute_name is None:
        label_attribute_name = name
    params = {
        'LabelingJobName': name,
        'LabelAttributeName': label_attribute_name,
        'InputConfig': _input_config(manifest_uri, free_of_pii, free_of_adult_content),
        'OutputConfig': _output_config(output_uri),
        'RoleArn': role_arn,
        'HumanTaskConfig': task_config
    }
    if category_config_uri:
        params['LabelCategoryConfigS3Uri'] = category_config_uri
    if algorithms_config:
        params['LabelingJobAlgorithmsConfig'] = algorithms_config
    if stopping_conditions:
        params['StoppingConditions'] = stopping_conditions
    return sagemaker_client.create_labeling_job(**params)


def describe_job(name, sagemaker_client=None):
    sagemaker_client = sagemaker_client if sagemaker_client else client
    return sagemaker_client.describe_labeling_job(LabelingJobName=name)


def get_job_state(name, sagemaker_client=None):
    sagemaker_client = sagemaker_client if sagemaker_client else client
    response = describe_job(name, sagemaker_client)
    status = response['LabelingJobStatus']
    labeled = response['LabelCounters']['TotalLabeled']
    unlabeled = response['LabelCounters']['Unlabeled']
    failed = response['LabelCounters']['FailedNonRetryableError']
    fail_message = ' {} failed'.format(failed) if failed > 0 else ''
    return "{} ({}/{})".format(status, labeled, unlabeled + labeled)+fail_message


def get_worker_responses(output_uri, job_name):
    by_worker = {}
    by_item = {}
    bucket_name, k = s3.decompose_uri(output_uri)
    for response_key in s3.list_objects(uri=os.path.join(output_uri, job_name, 'annotations/worker-response')):
        response_obj = s3.read_dict(bucket_name, response_key)
        item_id = response_key.split('/')[-2]
        by_item[item_id] = response_obj
        for response in response_obj['answers']:
            worker_id = response['workerId']
            responses = by_worker.get(worker_id, [])
            response['itemId'] = item_id
            responses.append(response)
            by_worker[worker_id] = responses
    return by_item, by_worker


def get_results(output_uri, job_name):
    try:
        return s3.read_list_of_dict(uri=os.path.join(output_uri, job_name, 'manifests/output/output.manifest'))
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchKey':
            return []
        else:
            raise


def get_multiple_results(output_uri, job_names, rename_label_to=None, exclude_failures=True):
    cumulative_results = []
    for job_name in job_names:
        results = get_results(output_uri, job_name)
        for item in results:
            if item[job_name + '-metadata'].get('failure-reason') is None or exclude_failures is False:
                if rename_label_to:
                    item[rename_label_to] = item.pop(job_name)
                    item[rename_label_to + '-metadata'] = item.pop(job_name + '-metadata')
                cumulative_results.append(item)
    return cumulative_results


def find_failures(manifest, attribute_name):
    failures = []
    reasons = {}
    for item in manifest:
        failure_reason = item[attribute_name+'-metadata'].get('failure-reason')
        if failure_reason:
            failures.append(item)
            failure_reason = failure_reason.replace(item['source-ref'], '<file>')
            failure_reason = failure_reason.replace(item['source-ref'].replace('/', '\\/'), '<file>')
            cnt = reasons.get(failure_reason,0)
            reasons[failure_reason] = cnt + 1
    return failures, reasons


def built_in_pre_lambda_bounding_box(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('PRE', region, 'BoundingBox')


def built_in_pre_lambda_image_multi_class(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('PRE', region, 'ImageMultiClass')


def built_in_pre_lambda_semantic_segmentation(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('PRE', region, 'SemanticSegmentation')


def built_in_pre_lambda_text_multi_class(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('PRE', region, 'TextMultiClass')


def built_in_pre_lambda_named_entity_recognition(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('PRE', region, 'NamedEntityRecognition')


def built_in_acs_lambda_bounding_box(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('ACS', region, 'BoundingBox')


def built_in_acs_lambda_image_multi_class(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('ACS', region, 'ImageMultiClass')


def built_in_acs_lambda_semantic_segmentation(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('ACS', region, 'SemanticSegmentation')


def built_in_acs_lambda_text_multi_class(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('ACS', region, 'TextMultiClass')


def built_in_acs_lambda_named_entity_recognition(region=None):
    region = __session.region_name if region is None else region
    return _built_in_lambda('ACS', region, 'NamedEntityRecognition')


def _built_in_lambda(mode, region, task):
    accounts = {
        'us-east-1': '432418664414',
        'us-east-2': '266458841044',
        'us-west-2': '081040173940',
        'ca-central-1': '918755190332',
        'eu-west-1': '568282634449',
        'eu-west-2': '487402164563',
        'eu-central-1': '203001061592',
        'ap-northeast-1':'477331159723',
        'ap-northeast-2':'845288260483',
        'ap-south-1':'565803892007',
        'ap-southeast-1':'377565633583',
        'ap-southeast-2':'454466003867'
    }
    account_id = accounts.get(region)
    if account_id:
        return 'arn:aws:lambda:{}:{}:function:{}-{}'.format(region, account_id, mode.upper(), task)
    else:
        raise Exception('Unrecognized region')


def scale_oversized_images_in_manifest(manifest, bucket=None, key_prefix=None, uri_prefix=None):
    new_manifest = []
    for item in manifest:
        new_item = item.copy()
        img, scalar = scale_image_to_size(uri=new_item['source-ref'])
        if scalar is not None:
            if uri_prefix:
                (bucket, key_prefix) = s3.decompose_uri(uri_prefix)
            if key_prefix is None:
                key_prefix = 'labeling_temp_images/'
            uri = s3.write_temp_object(img, key_prefix, bucket=bucket)
            new_item['original-source-ref'] = new_item.pop('source-ref')
            new_item['source-ref'] = uri
            new_item['scalar'] = scalar
        new_manifest.append(new_item)
    return new_manifest


def reverse_scaling_of_annotation(manifest, label_attribute_name, delete_scaled_images=True):
    new_manifest = []
    for item in manifest:
        new_item = item.copy()
        if 'scalar' in new_item:
            source_image = new_item.pop('old-source-ref')
            scalar = new_item.pop('scalar')
            scaled_image = new_item['source-ref']
            if delete_scaled_images:
                s3.delete_object(uri=scaled_image)
            new_item['source-ref'] = source_image
            for annotation in new_item[label_attribute_name]['annotations']:
                annotation['width'] = int(annotation['width'] / scalar)
                annotation['height'] = int(annotation['height'] / scalar)
                annotation['top'] = int(annotation['top'] / scalar)
                annotation['left'] = int(annotation['left'] / scalar)
        new_manifest.append(new_item)
    return new_manifest



import collections


class Box(collections.UserDict):
    def __init__(self, *data):
        collections.UserDict.__init__(self)
        for dat in data:
            self.update(dat)

    @classmethod
    def from_position(cls, position, calc_coordinates=False, **kwargs):
        obj = cls(position, kwargs)
        if calc_coordinates:
            obj._calc_coordinates()
        return obj

    @classmethod
    def from_coords(cls, coordinates, calc_position=False, **kwargs):
        obj = cls({'coordinates': coordinates}, kwargs)
        if calc_position:
            obj._calc_position()
        return obj

    def _calc_coordinates(self):
        self['coordinates'] = [self['left'], self['top'], self['left'] + self['width'] - 1,
                               self['top'] + self['height'] - 1]

    def _calc_position(self):
        self['left'] = self['coordinates'][0]
        self['top'] = self['coordinates'][1]
        self['width'] = self['coordinates'][2] - self['coordinates'][0] + 1
        self['height'] = self['coordinates'][3] - self['coordinates'][1] + 1

    def __missing__(self, key):
        if key == 'coordinates':
            self._calc_coordinates()
            return self['coordinates']
        else:
            raise KeyError(key)


import boto3
from botocore.exceptions import ClientError
import xml.etree.ElementTree as ET
import json
import zlib
import base64
import larry.s3 as s3
import larry.utils
from enum import Enum
from collections.abc import Iterable
from larry.mturk.HIT import HIT
from larry.mturk.Assignment import Assignment
import datetime


# Local client
client = None
# Indicate if we are working in production or sandbox
__production = True
# A local instance of the boto3 session to use
__session = boto3.session.Session()

PRODUCTION = 'production'
SANDBOX = 'sandbox'


def set_session(aws_access_key_id=None, aws_secret_access_key=None, aws__session_token=None,
                region_name=None, profile_name=None, boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws__session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session, client
    __session = boto_session if boto_session else boto3.session.Session(**larry.utils.copy_non_null_keys(locals()))
    s3.set_session(boto_session=__session)
    __create_client()


def __create_client():
    """
    Helper function to build an MTurk client. The client will use the current environment in use by the library,
    production or sandbox.
    """
    global client, __production, __session
    endpoint = "https://mturk-requester.us-east-1.amazonaws.com" if __production \
        else "https://mturk-requester-sandbox.us-east-1.amazonaws.com"
    client = __session.client(
        service_name='mturk',
        region_name='us-east-1',
        endpoint_url=endpoint
    )
    client.production = __production


def use_production():
    """
    Indicate that the library should use the production MTurk environment.
    :return: None
    """
    global __production, client
    __production = True
    __create_client()


def use_sandbox():
    """
    Indicate that the library should use the sandbox MTurk environment.
    :return: None
    """
    global __production, client
    __production = False
    __create_client()


def set_environment(environment=PRODUCTION, hit_id=None):
    """
    Set the environment that the library should use. If a HITId is provided, this will attempt to find
    the HIT in either environment and set the environment where that HITId is present.
    :param environment: The MTurk environment that the library should use (production or sandbox)
    :param hit_id: If provided, set the environment based on where the hit_id is present
    :return: If a hit_id is provided then the HIT response is returned, else None
    """
    def _try_get_hit(_hit_id):
        try:
            return get_hit(_hit_id)
        except ClientError as e:
            return e

    global client, __production, __session
    if hit_id:
        existing_production = __production

        # Try prod first
        if not __production:
            __production = True
            __create_client()
        hit = _try_get_hit(hit_id)
        if isinstance(hit, HIT):
            return hit
        else:
            __production = False
            __create_client()
            sandbox_hit = _try_get_hit(hit_id)
            if isinstance(sandbox_hit, HIT):
                return sandbox_hit
            else:
                __production = existing_production
                raise hit

    elif environment.lower() == SANDBOX:
        __production = False
        __create_client()
    else:
        __production = True
        __create_client()


def mturk_client_environment(c):
    if hasattr(c, 'production'):
        return c.production
    else:
        return c._endpoint.host == 'https://mturk-requester.us-east-1.amazonaws.com'


def production():
    return __production


def sandbox():
    return not __production


def environment():
    return PRODUCTION if __production else SANDBOX


# TODO Review how someone would use this to see if supporting iterables or other object types makes sense
def accept_qualification_request(request_id, value=None, mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    params = larry.utils.map_parameters(locals(), {'request_id': 'QualificationRequestId', 'value': 'IntegerValue'})
    mturk_client.accept_qualification_request(**params)


def approve(assignment, feedback=None, override_rejection=None, mturk_client=None):
    def _approve_assignment(_assignment_id, _feedback, _override_rejection):
        params = larry.utils.map_parameters(locals(), {
            '_assignment_id': 'AssignmentId',
            '_feedback': 'RequesterFeedback',
            '_override_rejection': 'OverrideRejection'
        })
        mturk_client.approve_assignment(**params)

    mturk_client = mturk_client if mturk_client else client
    if isinstance(assignment, str):
        _approve_assignment(assignment, feedback, override_rejection)
    elif hasattr(assignment, 'AssignmentId'):
        _approve_assignment(assignment['AssignmentId'], feedback, override_rejection)
    else:
        if isinstance(feedback, Iterable):
            for assignment_obj, fbk in zip(assignment, feedback):
                approve(assignment_obj, fbk, override_rejection, mturk_client)
        else:
            for assigment_obj in assignment:
                approve(assigment_obj, feedback, override_rejection, mturk_client)


approve_assignment = approve


def assign_qualification(qualification, worker_id, value=None, send_notification=None, mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    qualification_type_id = qualification['QualificationTypeId'] if hasattr(qualification,
                                                                            'QualificationTypeId') else qualification

    if isinstance(worker_id, Iterable):
        if isinstance(value, Iterable):
            for worker, val in zip(worker_id, value):
                assign_qualification(qualification_type_id, worker, val, send_notification, mturk_client)
        else:
            for worker in worker_id:
                assign_qualification(qualification_type_id, worker, value, send_notification, mturk_client)
    else:
        params = larry.utils.map_parameters(locals(), {
            'worker_id': 'WorkerId',
            'value': 'IntegerValue',
            'send_notification': 'SendNotification'
        })
        mturk_client.associate_qualification_with_worker(QualificationTypeId=qualification_type_id, **params)


associate_qualification_with_worker = assign_qualification


def add_assignments(hit_id, additional_assignments, request_token=None, mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    params = larry.utils.map_parameters(locals(), {
        'hit_id': 'HITId',
        'additional_assignments': 'NumberOfAdditionalAssignments',
        'request_token': 'UniqueRequestToken'
    })
    mturk_client.create_additional_assignments_for_hit(**params)


create_additional_assignments_for_hit = add_assignments


def create_hit(title,
               description,
               reward=None,
               reward_cents=None,
               lifetime=86400,
               assignment_duration=3600,
               max_assignments=None,
               auto_approval_delay=None,
               keywords=None,
               question=None,
               annotation=None,
               qualification_requirements=None,
               request_token=None,
               assignment_review_policy=None,
               hit_review_policy=None,
               hit_layout_id=None,
               hit_layout_parameters=None,
               mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    params = larry.utils.map_parameters(locals(), {
        'title': 'Title',
        'description': 'Description',
        'reward': 'Reward',
        'lifetime': 'LifetimeInSeconds',
        'assignment_duration': 'AssignmentDurationInSeconds',
        'max_assignments': 'MaxAssignments',
        'auto_approval_delay': 'AutoApprovalDelayInSeconds',
        'keywords': 'Keywords',
        'question': 'Question',
        'annotation': 'RequesterAnnotation',
        'qualification_requirements': 'QualificationRequirements',
        'request_token': 'UniqueRequestToken',
        'assignment_review_policy': 'AssignmentReviewPolicy',
        'hit_review_policy': 'HITReviewPolicy',
        'hit_layout_id': 'HITLayoutId',
        'hit_layout_parameters': 'HITLayoutParameters'
    })
    if reward_cents:
        params['reward'] = str(reward_cents / 100)
    return HIT(mturk_client.create_hit(**params).get('HIT'))


def create_hit_type(title,
                    description,
                    reward,
                    lifetime,
                    assignment_duration,
                    max_assignments=None,
                    auto_approval_delay=None,
                    keywords=None,
                    qualification_requirements=None,
                    mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    params = larry.utils.map_parameters(locals(), {
        'title': 'Title',
        'description': 'Description',
        'reward': 'Reward',
        'lifetime': 'LifetimeInSeconds',
        'assignment_duration': 'AssignmentDurationInSeconds',
        'max_assignments': 'MaxAssignments',
        'auto_approval_delay': 'AutoApprovalDelayInSeconds',
        'keywords': 'Keywords',
        'qualification_requirements': 'QualificationRequirements'
    })
    return mturk_client.create_hit_type(**params).get('HITTypeId')


def _get_assignment(assignment_id, mturk_client=None):
    """
    Retrieves the Assignment and HIT data for a given AssignmentID. The Assignment data is updated to replace the
    Answer XML with a dict object.
    :param assignment_id: The assignment to retrieve
    :param mturk_client: The MTurk client to use for this request instead of the default client
    :return: A tuple of assignment and hit dicts
    """
    mturk_client = mturk_client if mturk_client else client
    response = mturk_client.get_assignment(AssignmentId=assignment_id)
    return response['Assignment'], response['HIT']


def get_assignment(assignment_id, mturk_client=None):
    """
    Retrieves the Assignment and HIT data for a given AssignmentID. The Assignment data is updated to replace the
    Answer XML with a dict object.
    :param assignment_id: The assignment to retrieve
    :param mturk_client: The MTurk client to use for this request instead of the default client
    :return: A tuple of assignment and hit dicts
    """
    a, h = _get_assignment(assignment_id, mturk_client)
    return Assignment(a), HIT(h)


def get_account_balance(mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    return float(mturk_client.get_account_balance()['AvailableBalance'])


def _get_hit(hit_id, mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    return mturk_client.get_hit(HITId=hit_id)['HIT'], mturk_client_environment(mturk_client)


def get_hit(hit_id, mturk_client=None):
    """
    Retrieve HIT data for the id
    :param hit_id: ID to retrieve
    :param mturk_client: The MTurk client to use for this request instead of the default client
    :return: A dict containing HIT attributes
    """
    hit, prod = _get_hit(hit_id, mturk_client)
    return HIT(hit, production=prod)


def list_assignments_for_hit(hit_id, submitted=True, approved=True, rejected=True, mturk_client=None):
    """
    Retrieves all of the assignments for a HIT with the Answer XML processed into a dict.
    :param hit_id: ID of the HIT to retrieve
    :param submitted: Boolean indicating if assignments in a state of Submitted should be retrieved, default is True
    :param approved: Boolean indicating if assignments in a state of Approved should be retrieved, default is True
    :param rejected: Boolean indicating if assignments in a state of Rejected should be retrieved, default is True
    :param mturk_client: The MTurk client to use for this request instead of the default client
    :return: A generator containing the assignments
    """
    if mturk_client is None:
        mturk_client = client
    statuses = []
    if submitted:
        statuses.append('Submitted')
    if approved:
        statuses.append('Approved')
    if rejected:
        statuses.append('Rejected')
    pages_to_get = True
    next_token = None
    while pages_to_get:
        if next_token:
            response = mturk_client.list_assignments_for_hit(HITId=hit_id, NextToken=next_token,
                                                             AssignmentStatuses=statuses)
        else:
            response = mturk_client.list_assignments_for_hit(HITId=hit_id, AssignmentStatuses=statuses)
        if response.get('NextToken'):
            next_token = response['NextToken']
        else:
            pages_to_get = False
        for assignment in response.get('Assignments', []):
            yield Assignment(assignment)


def list_hits(mturk_client=None):
    """
    Retrieves all of the HITs in your account with the exception of those that have been deleted (automatically or by
    request).
    :param mturk_client: The MTurk client to use for this request instead of the default client
    :return: A generator containing the HITs
    """
    if mturk_client is None:
        mturk_client = client
    pages_to_get = True
    next_token = None
    while pages_to_get:
        if next_token:
            response = mturk_client.list_hits(NextToken=next_token)
        else:
            response = mturk_client.list_hits()
        if response.get('NextToken'):
            next_token = response['NextToken']
        else:
            pages_to_get = False
        for hit in response.get('HITs', []):
            yield HIT(hit)


def create_qualification_type(name, description, keywords=None, status='Active', retry_delay=None, test=None,
                              test_duration=None, answer_key=None, auto_granted=False, auto_granted_value=None,
                              mturk_client=None):
    if mturk_client is None:
        mturk_client = client
    params = larry.utils.map_parameters(locals(), {
        'name': 'Name',
        'keywords': 'Keywords',
        'description': 'Description',
        'status': 'QualificationTypeStatus',
        'retry_delay': 'RetryDelayInSeconds',
        'test': 'Test',
        'answer_key': 'AnswerKey',
        'test_duration': 'TestDurationInSeconds',
        'auto_granted': 'AutoGranted',
        'auto_granted_value': 'AutoGrantedValue'
    })
    response = mturk_client.create_qualification_type(**params)
    return response['QualificationType']['QualificationTypeId']


# TODO: Add methods for multiple applications
def assign_qualification(qualification_type_id, worker_id, value=None, send_notification=False, mturk_client=None):
    if mturk_client is None:
        mturk_client = client
    params = larry.utils.map_parameters(locals(), {
        'qualification_type_id': 'QualificationTypeId',
        'worker_id': 'WorkerId',
        'send_notification': 'SendNotification',
    })
    if value:
        params['IntegerValue'] = value
    return mturk_client.associate_qualification_with_worker(**params)


# TODO: Add methods for multiple applications
def remove_qualification(qualification_type_id, worker_id, reason=None, mturk_client=None):
    if mturk_client is None:
        mturk_client = client
    params = larry.utils.map_parameters(locals(), {
        'qualification_type_id': 'QualificationTypeId',
        'worker_id': 'WorkerId',
        'reason': 'Reason'
    })
    return mturk_client.disassociate_qualification_from_worker(**params)


def preview_url(hit_type_id, production=None):
    """
    Generates a preview URL for the hit_type_id using the current environment.
    :param hit_type_id: The HIT type
    :param production: True if the request is for the production environment
    :return: The preview URL
    """
    global __production
    prod = __production if __production is None else __production
    if prod:
        return "https://worker.mturk.com/mturk/preview?groupId={}".format(hit_type_id)
    else:
        return "https://workersandbox.mturk.com/mturk/preview?groupId={}".format(hit_type_id)


# TODO: Create an enum for destination types
def add_notification(hit_type_id, destination, event_types, mturk_client=None):
    """
    Attaches a notification to the HIT type to send a message when various event_types occur
    :param hit_type_id: The HIT type to attach a notification to
    :param destination: An SNS ARN or a SQS URL
    :param event_types: A list of event types to trigger messages on; valid types are:
    :param mturk_client: The MTurk client to use for this request instead of the default client
    AssignmentAccepted | AssignmentAbandoned | AssignmentReturned | AssignmentSubmitted | AssignmentRejected |
    AssignmentApproved | HITCreated | HITExtended | HITDisposed | HITReviewable | HITExpired | Ping
    :return: The API response
    """
    if mturk_client is None:
        mturk_client = client
    return mturk_client.update_notification_settings(
        HITTypeId=hit_type_id,
        Notification={
            'Destination': destination,
            'Transport': 'SQS' if 'sqs' in destination else 'SNS',
            'Version': '2014-08-15',
            'EventTypes': event_types
        },
        Active=True
    )


def update_expiration(hit_id, expire_at, mturk_client=None):
    mturk_client = mturk_client if mturk_client else client
    mturk_client.update_expiration_for_hit(HITId=hit_id, ExpireAt=expire_at)


def expire_hit(hit_id, mturk_client=None):
    update_expiration(hit_id, datetime.datetime.now(), mturk_client)


def parse_answers(answer):
    """
    Parses the answer XML into a usable python dict for analysis. In cases where answers contain JSON strings
    it attempts to parse them into dicts. Does not support file upload answers.
    :param answer: An MTurk Answer object
    :return: A dict containing the parsed answer data
    """
    result = {}
    ns = {'mt': 'http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2005-10-01/QuestionFormAnswers.xsd'}
    root = ET.fromstring(answer)

    answers = root.findall('mt:Answer', ns)

    # TODO: Need to test this for the various types
    # if this is coming from the crowd-form element load the JSON
    if len(answers) == 1 and answers[0].find('mt:QuestionIdentifier', ns).text == 'taskAnswers':
        answer_body = json.loads(answers[0].find('mt:FreeText', ns).text)

        # For some reason crowd form creates what appears to be an unnecessary list, in those cases we flatten it
        if isinstance(answer_body, list) and len(answer_body) == 1:
            answer_body = answer_body[0]
            for key, value in answer_body.items():
                if isinstance(value, str):
                    try:
                        result[key] = json.loads(value)
                    except json.decoder.JSONDecodeError:
                        result[key] = value
                else:
                    result[key] = value
        else:
            result['taskAnswers'] = answer_body
    # else this is a standard element and each value should be loaded
    else:
        for a in answers:
            name = a.find('mt:QuestionIdentifier', ns).text
            if a.find('mt:FreeText', ns) is not None:
                answer_text = a.find('mt:FreeText', ns).text
                try:
                    result[name] = json.loads(answer_text)
                except json.decoder.JSONDecodeError:
                    result[name] = answer_text
            else:
                selection_elements = a.findall('mt:SelectionIdentifier')
                selections = []
                for selection in selection_elements:
                    selections.append(selection.text)
                if a.find('mt:OtherSelection'):
                    selections.append(a.find('mt:OtherSelection').text)
                result[name] = selections
    return result


def prepare_requester_annotation(payload, s3_resource=None, bucket_identifier=None):
    """
    Converts content into a format that can be inserted into the RequesterAnnotation field of a HIT when it is
    created. Because the RequesterAnnotation field is limited to 255 characters this will automatically format it
    as efficiently as possible depending on the size of the content. If the content is short enough the payload will
    be retained as is. If it's too long, it will attempt to compress it using zlib. And if it's still too long it will
    be stored in a temporary file in S3.

    Note that using this may result in creating a '*larry*' bucket in your S3 environment which will require
    create-bucket permissions for your user. When retrieving the annotation you have the option to request that any
    temp files be deleted.
    :param payload: The content to be stored in the RequesterAnnotation field
    :param s3_resource: The s3 resource to use for writing to S3 if you don't want to use the default one
    :param bucket_identifier: The identifier to attach to the temp bucket that will be used for writing to s3, typically
    the account id (from STS) for the account being used
    :return: A string value that can be placed in the RequesterAnnotation field
    """
    s3_resource = s3_resource if s3_resource else s3.resource
    payload_string = json.dumps(payload, separators=(',', ':')) if type(payload) == dict else payload

    # Use the annotation 'as is' if possible
    if len(payload_string) < 243:
        return json.dumps({'payload': payload}, separators=(',', ':'))

    else:
        # Attempt to compress it
        compressed = str(base64.b85encode(zlib.compress(payload_string.encode())), 'utf-8')
        if len(compressed) < 238:
            return json.dumps({'payloadBytes': compressed}, separators=(',', ':'))

        else:
            # Else post it to s3
            uri = s3.write_temp_object(payload, 'mturk_requester_annotation/', s3_resource=s3_resource,
                                       bucket_identifier=bucket_identifier)
            return json.dumps({'payloadURI': uri}, separators=(',', ':'))


def retrieve_requester_annotation(hit_id, delete_temp_file=False, s3_resource=None, mturk_client=None):
    """
    Takes a value from the RequesterAnnotation field that was stored by the prepare_requester_annotation function
    and extracts the relevant payload from the text, compressed bytes, or S3.
    :param hit_id: The ID of the HIT to retrieve the RequesterAnnotation for
    :param delete_temp_file: True if you wish to delete the payload S3 object if one was created.
    :param s3_resource: The S3 resource to use when retrieving annotations if necessary
    :param mturk_client: The MTurk client to use if you don't want to use the default client
    :return: The payload that was originally stored by prepare_requester_annotation
    """
    if mturk_client is None:
        mturk_client = client
    hit, p = _get_hit(hit_id, mturk_client=mturk_client)
    return parse_requester_annotation(hit.get('RequesterAnnotation', ''), delete_temp_file, s3_resource)


def parse_requester_annotation(content, delete_temp_file=False, s3_resource=None):
    s3_resource = s3_resource if s3_resource else s3.resource
    if content and len(content) > 0:
        try:
            content = json.loads(content)
            if 'payload' in content:
                return content['payload']
            elif 'payloadBytes' in content:
                return json.loads(zlib.decompress(base64.b85decode(content['payloadBytes'].encode())))
            elif 'payloadURI' in content:
                results = s3.read_dict(uri=content['payloadURI'], s3_resource=s3_resource)
                if delete_temp_file:
                    s3.delete_object(uri=content['payloadURI'], s3_resource=s3_resource)
                return results
            else:
                return content
        except json.decoder.JSONDecodeError:
            return content
    else:
        return content


def render_jinja_template(arguments, template=None, template_uri=None):
    """
    Renders an HTML question task using jinja2 to populate a provided template with the arguments provided.
    :param arguments: A dict containing the values to use in rendering the template. Requires Jinja2 be installed.
    :param template: A string value for the template
    :param template_uri: The URI of an S3 object to use as a template
    :return: The contents of the rendered template
    """
    try:
        from jinja2 import Template
        if template_uri:
            template = s3.read_str(uri=template_uri)
        jinja_template = Template(template)
        return jinja_template.render(arguments)
    except ImportError as e:
        # We'll simply raise the ImportError to let the developer know this requires Jinja2 to function
        raise e


def render_jinja_template_question(arguments, template=None, template_uri=None):
    """
    Renders an HTML question task using jinja2 to populate a provided template with the arguments provided.
    :param arguments: A dict containing the values to use in rendering the template. Requires Jinja2 be installed.
    :param template: A string value for the template
    :param template_uri: The URI of an S3 object to use as a template
    :return: The HTMLQuestion XML to use for the task
    """
    return render_html_question(render_jinja_template(arguments, template, template_uri))


def render_html_question(html, frame_height=0):
    """
    Renders HTML task content within an HTMLQuestion XML object for use as a task.
    :param html: HTML string to render within the template
    :param frame_height: Frame height to use for the Worker viewport, zero by default to use the whole window
    :return: The rendered HTMLQuestion XML string
    """
    return '''<HTMLQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2011-11-11/HTMLQuestion.xsd">
                <HTMLContent><![CDATA[{}]]></HTMLContent><FrameHeight>{}</FrameHeight>
              </HTMLQuestion>'''.format(html, frame_height)


def render_external_question(url, frame_height=0):
    """
    Renders a URL within an ExternalQuestion XML object for use as a task.
    :param url: The URL of the task to display to Workers
    :param frame_height: Frame height to use for the Worker viewport, zero by default to use the whole window
    :return: The rendered ExternalQuestion XML string
    """
    return '''<ExternalQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd">
                <ExternalURL>{}</ExternalURL><FrameHeight>{}</FrameHeight>
              </ExternalQuestion>'''.format(url, frame_height)


def list_sns_events(event, event_filter=None):
    for record in event['Records']:
        message_id = record['Sns']['MessageId']
        notification = json.loads(record['Sns']['Message'])
        for mturk_event in notification['Events']:
            if event_filter is None or mturk_event['EventType'] == event_filter:
                yield mturk_event, message_id


def list_sqs_events(event, event_filter=None):
    for record in event['Records']:
        message_id = record['MessageId']
        notification = json.loads(record['body'])
        for mturk_event in notification['Events']:
            if event_filter is None or mturk_event['EventType'] == event_filter:
                yield mturk_event, message_id


class QualificationComparitor(Enum):
    LessThan = "LessThan"
    LessThanOrEqualTo = "LessThanOrEqualTo"
    GreaterThan = "GreaterThan"
    GreaterThanOrEqualTo = "GreaterThanOrEqualTo"
    EqualTo = "EqualTo"
    NotEqualTo = "NotEqualTo"
    Exists = "Exists"
    DoesNotExist = "DoesNotExist"
    In = "In"
    NotIn = "NotIn"


class QualificationActionsGuarded(Enum):
    Accept = "Accept"
    PreviewAndAccept = "PreviewAndAccept"
    DiscoverPreviewAndAccept = "DiscoverPreviewAndAccept"


def build_qualification_requirement(qualification_type_id, comparator, values=None, value=None,
                                    locales=None, locale=None, actions_guarded=None):
    if value:
        values = [value]
    if locale:
        locales = [locale]
    requirement = {
        'QualificationTypeId': qualification_type_id,
        'Comparator': comparator.value if isinstance(comparator, QualificationComparitor) else comparator,
    }
    if values:
        requirement['IntegerValues'] = values
    if locales:
        _locales = []
        for locale in locales:
            if isinstance(locale, tuple):
                _locales.append({'Country': locale[0], 'Subdivision': locale[1]})
            else:
                _locales.append({'Country': locale})
        requirement['LocaleValues'] = _locales
    if actions_guarded:
        if isinstance(actions_guarded, QualificationActionsGuarded):
            requirement['ActionsGuarded'] = actions_guarded.value
        else:
            requirement['ActionsGuarded'] = actions_guarded
    return requirement


def build_masters_requirement(actions_guarded=None):
    if __production:
        return build_qualification_requirement('2F1QJWKUDD8XADTFD2Q0G6UTO95ALH',
                                               QualificationComparitor.Exists,
                                               actions_guarded=actions_guarded)
    else:
        return build_qualification_requirement('2ARFPLSP75KLA8M8DH1HTEQVJT3SY6',
                                               QualificationComparitor.Exists,
                                               actions_guarded=actions_guarded)


def build_adult_requirement(actions_guarded=None):
    return build_qualification_requirement('00000000000000000060',
                                           QualificationComparitor.EqualTo,
                                           value=1,
                                           actions_guarded=actions_guarded)


def build_hits_approved_requirement(comparator, value, actions_guarded=None):
    return build_qualification_requirement('00000000000000000040',
                                           comparator,
                                           value=value,
                                           actions_guarded=actions_guarded)


def build_percent_approved_requirement(comparator, value, actions_guarded=None):
    return build_qualification_requirement('000000000000000000L0',
                                           comparator,
                                           value=value,
                                           actions_guarded=actions_guarded)


def build_locale_requirement(comparator, locales=None, locale=None, actions_guarded=None):
    if locale:
        locales = [locale]
    if locales is None or len(locales) == 0:
        raise Exception('A locale qualification requires at least one locale')
    return build_qualification_requirement('00000000000000000071',
                                           comparator=comparator,
                                           locales=locales,
                                           actions_guarded=actions_guarded)


# MTurk utility functions
def retrieve_hit_results_to_dict(items, hit_id_attribute='HITId', assignments_attribute='Assignments'):
    stats = {}
    for item in items:
        if hit_id_attribute in item:
            hit_id = item[hit_id_attribute]
            hit = get_hit(hit_id)
            status = hit['HITStatus']
            item['HITStatus'] = status
            stats[status] = stats.get(status, 0) + 1
            item[assignments_attribute] = list(list_assignments_for_hit(hit_id))
            completed = len(item[assignments_attribute])
            stats['CompletedAssignments'] = stats.get('CompletedAssignments', 0) + completed
            stats['MaxAssignments'] = stats.get('MaxAssignments', 0) + hit['MaxAssignments']
            if completed < hit['MaxAssignments'] and status == 'Reviewable':
                stats['ExpiredAssignments'] = stats.get('ExpiredAssignments', 0) - completed + hit['MaxAssignments']
    return stats



import larry.s3 as s3
import larry.mturk as mturk
import larry.sqs as sqs
import larry.sts as sts
import larry.sagemaker as sagemaker
import boto3
import larry.utils as utils
import larry.types as types
import json

__version__ = "0.1.4"


def _propagate_session():
    s3.set_session(boto_session=__session)
    mturk.set_session(boto_session=__session)
    sqs.set_session(boto_session=__session)
    sts.set_session(boto_session=__session)
    sagemaker.set_session(boto_session=__session)


# A local instance of the boto3 session to use
__session = boto3.session.Session()
_propagate_session()


def session():
    global __session
    if __session is None:
        __session = boto3.session.Session()
    return __session


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws_session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws_session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session
    __session = boto_session if boto_session is not None else boto3.session.Session(**utils.copy_non_null_keys(locals()))
    _propagate_session()




import json
import datetime
import larry

DATE_FORMAT = '%Y-%m-%d %H:%M:%S%z'


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return date_to_string(obj)
        elif isinstance(obj, larry.mturk.HIT):
            hit = {i: obj[i] for i in obj if i != 'Question'}
            hit['__HIT__'] = True
            return hit
        elif isinstance(obj, larry.mturk.Assignment):
            assignment = {i: obj[i] for i in obj}
            assignment['__Assignment__'] = True
            return assignment
        return json.JSONEncoder.default(self, obj)


def JSONDecoder(dct):
    if '__HIT__' in dct:
        return larry.mturk.HIT(dct)
    if '__Assignment__' in dct:
        return larry.mturk.Assignment(dct)
    return dct


def date_to_string(obj):
    return obj.strftime(DATE_FORMAT)


def parse_date(obj):
    return datetime.datetime.strptime(obj, DATE_FORMAT)


def copy_non_null_keys(param_list):
    result = {}
    for key, val in param_list.items():
        if val is not None:
            result[key] = val
    return result


def map_parameters(parameters, key_map):
    result = {}
    for k, i in key_map.items():
        if parameters.get(k) is not None:
            result[i] = parameters[k]
    return result



import unittest
import larry as ld

# S3 testing objects
my_dict = {
    'a': {'key': 'value'},
    'b': ['a', 'b', 'c'],
    '1': 15
}
my_listofdicts = [
    {'a': 124, 'b': 'value'},
    {'a': 939, 'b': 'foo'},
    {'a': 389, 'b': 'bar', '3': 'new'}
]
my_list = ['a', 'b', 'c', '3', 'd', '2', 'e', '1']
my_string = 'foobar'
my_image_url = 'https://hilltop-demo.s3-us-west-2.amazonaws.com/images/1557026914963.jpg'
prefix = 'larry-testing/'


class S3Tests(unittest.TestCase):

    def test_readwrite_dict(self):
        dict_uri = ld.s3.write_temp_object(my_dict, prefix)
        self.assertEqual(ld.s3.read_dict(uri=dict_uri),  my_dict)

    def test_readwrite_list_of_dict(self):
        listofdicts_uri = ld.s3.write_temp_object(my_listofdicts, prefix)
        self.assertEqual(ld.s3.read_list_of_dict(uri=listofdicts_uri), my_listofdicts)

    def test_readwrite_list(self):
        list_uri = ld.s3.write_temp_object(my_list, prefix)
        self.assertEqual(ld.s3.read_list_of_str(uri=list_uri), my_list)

    def test_readwrite_string(self):
        string_uri = ld.s3.write_temp_object(my_string, prefix)
        self.assertEqual(ld.s3.read_str(uri=string_uri), my_string)

    def test_rename_object(self):
        dict_uri = ld.s3.write_temp_object(my_dict, prefix)
        temp_bucket, src_key = ld.s3.decompose_uri(dict_uri)
        ld.s3.rename_object(temp_bucket, src_key, temp_bucket, src_key + '.renamed')
        new_uri = ld.s3.compose_uri(temp_bucket, src_key + '.renamed')
        self.assertEqual(new_uri, dict_uri+'.renamed')

    def test_object_exists(self):
        dict_uri = ld.s3.write_temp_object(my_dict, prefix)
        self.assertTrue(ld.s3.object_exists(uri=dict_uri))

    def test_list_and_delete(self):
        dict_uri = ld.s3.write_temp_object(my_dict, prefix)
        temp_bucket, src_key = ld.s3.decompose_uri(dict_uri)
        for key in ld.s3.list_objects(temp_bucket, prefix):
            ld.s3.delete_object(temp_bucket, key)
        objects = list(ld.s3.list_objects(temp_bucket, prefix))
        self.assertEqual(len(objects),0)

    def test_fetch(self):
        fetched_uri = ld.s3.fetch(my_image_url, bucket=temp_bucket, key=prefix + 'fetched.jpg')
        image = ld.s3.read_pillow_image(uri=fetched_uri)
        image_uri = ld.s3.write_pillow_image(image, 'JPEG', uri=fetched_uri + '.rewritten.jpg')
        image2 = ld.s3.read_pillow_image(uri=image_uri)
        ld.s3.download_to_temp(uri=image_uri)
        image_url = ld.s3.make_public(uri=image_uri)
        fetched_uri2 = ld.s3.fetch(image_url, bucket=temp_bucket, key=prefix + 'fetched2.jpg')
        image3 = ld.s3.read_pillow_image(uri=fetched_uri2)
        ld.s3.delete_object(uri=fetched_uri)
        ld.s3.delete_object(uri=image_uri)
        ld.s3.delete_object(uri=fetched_uri2)
        image3

if __name__ == '__main__':
    unittest.main()





def _extract_response_detail(assignments, identifier, exclude_rejected=True):
    responses = []
    response_count = 0
    work_time = 0
    for assignment in assignments:
        responses.append({
            'WorkerId': assignment['WorkerId'],
            'Response': assignment['Answer'].get(identifier),
            'AssignmentId': assignment['AssignmentId'],
            'HITId': assignment['HITId'],
            'AcceptTime': assignment['AcceptTime'],
            'WorkTime': assignment['WorkTime'],
            'Excluded': assignment['AssignmentStatus'] == 'Rejected'
        })
        if exclude_rejected is False or assignment['AssignmentStatus'] != 'Rejected':
            response_count += 1
            work_time += int(assignment['WorkTime'].total_seconds())
    return {
        'Responses': responses,
        'ResponseCount': response_count,
        'WorkTime': work_time,
        'Identifier': identifier
    }


def _score_text_responses(response_detail):
    scores = {}
    for response in response_detail['Responses']:
        value = response['Response']
        score = scores.get(value, 0)
        scores[value] = score + 1
    return scores


def _consolidate_text_response(assignments, identifier, threshold, exclude_rejected=True):
    response_detail = _extract_response_detail(assignments, identifier, exclude_rejected=exclude_rejected)
    response_detail['ScoredResponses'] = _score_text_responses(response_detail)
    answer = None
    for response, score in response_detail['ScoredResponses'].items():
        if score * 100 / response_detail['ResponseCount'] >= threshold:
            answer = response
            break
    if answer is not None:
        for response in response_detail['Responses']:
            response['Accuracy'] = response['Response'] == answer
    return answer, response_detail


def consolidate_crowd_classifier(assignments, threshold=60, exclude_rejected=True):
    """
    Retrieves Worker responses for a HITId and computes a consolidated answer based on a simple plurality of responses.
    For example, if the HIT has 3 Assignments, and Workers respond with responses of A, A, and B, the resulting
    response would be A since 66.7% of Workers agree on a response of A which is higher than the default threshold of
    60%. If the threshold were set at 80% than None would be returned. Similarly, if Workers responded with A, B, and C,
    the result would be None since none of the answers received 60% of responses. By default, Assignments that have
    already been rejected are ignored for purposes of scoring responses.
    :param assignments: A list or generator containing the assignments to process
    :param threshold: A 0-100 percentage value (80 = 80%) to use a a threshold in looking for agreement amongst Workers
    :param exclude_rejected: Boolean value (default=True) indicating that Assignments that have already been
    rejected should be excluded
    :return: A tuple containing the result and an object with detail on the responses for use in measuring Worker
    accuracy
    """
    return _consolidate_text_response(
        assignments,
        'category.label',
        threshold,
        exclude_rejected=exclude_rejected)